"""
Factitious CA to test non-uniform buffer interactions.

Experiment classes included.

"""
from xentica import core
from xentica import seeds
from xentica.core import color_effects


class ShiftingSands(core.CellularAutomaton):
    """
    CA for non-uniform buffer interactions test.

    It emits the whole value to a constant direction, then absorbs
    surrounding values by summing them.

    """

    state = core.IntegerProperty(max_val=1)

    class Topology:
        """2D Moore neighborhood, wrapped to a 3-torus."""

        dimensions = 2
        lattice = core.OrthogonalLattice()
        neighborhood = core.MooreNeighborhood()
        border = core.TorusBorder()

    def emit(self):
        """Emit the whole value to a constant direction."""
        direction = 0
        for i in range(len(self.buffers)):
            if i == direction:
                self.buffers[i].state = self.main.state
            else:
                self.buffers[i].state = 0

    def absorb(self):
        """Absorb surrounding values by summing them."""
        new_val = core.IntegerVariable()
        for i in range(len(self.buffers)):
            new_val += self.neighbors[i].buffer.state
        self.main.state = new_val

    @color_effects.MovingAverage
    def color(self):
        """Render contrast black & white cells."""
        red = self.main.state * 255
        green = self.main.state * 255
        blue = self.main.state * 255
        return (red, green, blue)


class ShiftingSandsExperiment(core.Experiment):
    """Vanilla experiment with randomly initialized area."""

    word = "LET IT SHIFT"
    size = (640, 360, )
    zoom = 3
    pos = [0, 0]
    seed = seeds.patterns.BigBang(
        pos=(320, 180),
        size=(100, 100),
        vals={
            "state": seeds.random.RandInt(0, 1),
        }
    )


def main():
    """Run model/experiment interactively."""
    import moire
    model = ShiftingSands(ShiftingSandsExperiment)
    gui = moire.GUI(runnable=model)
    gui.run()


if __name__ == "__main__":
    main()



"""
The collection of classes to describe experiments for CA models.

Experiment is a class with CA parameters stored as class
variables. Different models may have a different set of parameters. To
make sure all set correct, you should inherit your experiments from
:class:`Experiment` class.

The quick example::

    from xentica import core, seeds

    class MyExperiment(core.Experiment):
        # RNG seed string
        word = "My Special String"
        # field size
        size = (640, 360, )
        # initial field zoom
        zoom = 3
        # initial field shift
        pos = [0, 0]
        # A pattern used in initial board state generation.
        # BigBang is a small area initialized with high-density random values.
        seed = seeds.patterns.BigBang(
            # position Big Bang area
            pos=(320, 180),
            # size of Big Bang area
            size=(100, 100),
            # algorithm to generate random values
            vals={
                "state": seeds.random.RandInt(0, 1),
            }
       )



"""

__all__ = ['Experiment', ]


class Experiment:
    """
    Base class for all experiments.

    Right now doing nothing, but will be improved in future
    versions. So it is adviced to inherit your experiments from it.

    """



"""
The Python CA engine.

Xentica is the framework to build GPU-accelerated models for
multi-dimensional cellular automata. Given pure Python definitions, it
generates kernels in CUDA C and runs them on NVIDIA hardware.

"""



"""
The collection of classes implementing render logic.

The renderer takes the array of cells' colors and renders the screen
frame from it. Also, it is possible to expand a list of user actions,
adding ones specific to the renderer, like zoom, scroll etc.

The default renderer is :class:`RendererPlain`. Though there are no
other renderers yet, you may try to implement your own and apply it to
CA model as follows::

    from xentica.core import CellularAutomaton
    from xentica.core.renderers import Renderer

    class MyRenderer(Renderer):
        # ...

    class MyCA(CellularAutomaton):
        renderer = MyRenderer()
        # ...

"""
import abc

import numpy as np

from xentica.core.mixins import BscaDetectorMixin

__all__ = ['Renderer', 'RendererPlain', ]


class Renderer(BscaDetectorMixin):
    """
    Base class for all renderers.

    For correct behavior, renderer classes should be inherited from
    this class. Then at least :meth:`render_code` method should be implemented.

    However, if you are planning to add user actions specific to your
    renderer, more methods should be overridden:

    - :meth:`__init__`, where you expand a list of kernel arguments in
      ``self.args``;

    - :meth:`get_args_vals`, where you expand the list of arguments' values;

    - :meth:`setup_actions`, where you expand a dictionary of bridge actions;

    See :class:`RendererPlain` code as an example.

    """

    def __init__(self):
        """Initialize kernel arguments."""
        self.args = [
            ("int3", "*col"),
            ("int", "*img"),
        ]

    @staticmethod
    def get_args_vals(bsca):
        """
        Get a list of kernel arguments values.

        The order should correspond to ``self.args``, with the values
        themselves as either PyCUDA ``GpuArray`` or correct NumPy
        instance. Those values will be used directly as arguments to
        PyCUDA kernel execution.

        :param bsca:
            :class:`xentica.core.CellularAutomaton` instance.

        """
        args_vals = [bsca.gpu.arrays.colors, bsca.gpu.arrays.img]
        return args_vals

    @abc.abstractmethod
    def render_code(self):
        """
        Generate C code for rendering.

        At minimum, it should process cells colors stored in ``col``
        GPU-array, and store the resulting pixel's value into ``img``
        GPU-array. It can additionally use other custom arguments, if
        any set up.

        """
        return ""

    def setup_actions(self, bridge):
        """
        Expand bridge with custom user actions.

        You can do it as follows::

            class MyRenderer(Renderer):
                # ...

                @staticmethod
                def my_awesome_action():
                    def func(ca, gui):
                        # do something with ``ca`` and ``gui``
                    return func

                def setup_actions(self):
                    bridge.key_actions.update({
                        "some_key": self.my_awesome_action(),
                    })

        :param bridge:
            :class:`xentica.bridge.Bridge` instance.

        """


class RendererPlain(Renderer):
    """
    Render board as 2D plain.

    If your model has more than 2 dimensions, a projection over
    ``projection_axes`` tuple will be made. The default is two first
    axes, which corresponds to ``(0, 1)`` tuple.

    """

    def __init__(self, projection_axes=None):
        """Initialize custom kernel arguments and projection axes.

        :param projection_axes:
            A tuple with indexes of 2 axes over which a projection is
            made.  If ``None`` value is given, two first axes, ``(0, 1)``
            will be used.

        """
        super(RendererPlain, self).__init__()
        self.args += [
            ("int", "zoom"),
            ("int", "dx"),
            ("int", "dy"),
            ("int", "width"),
        ]
        self.projection_axes = projection_axes
        if self.projection_axes is None:
            self.projection_axes = (0, 1)

    def get_args_vals(self, bsca):
        """Extend kernel arguments values."""
        args_vals = super(RendererPlain, self).get_args_vals(bsca)
        args_vals += [
            np.int32(bsca.zoom),
            np.int32(bsca.pos[0]),
            np.int32(bsca.pos[1]),
            np.int32(bsca.width),
        ]
        return args_vals

    def setup_actions(self, bridge):
        """Extend bridge with scroll and zoom user actions."""
        bridge.key_actions.update({
            "up": self.move(0, 1),
            "down": self.move(0, -1),
            "left": self.move(-1, 0),
            "right": self.move(1, 0),
            "=": self.zoom(1),
            "-": self.zoom(-1),
        })

    @staticmethod
    def move(delta_x, delta_y):
        """
        Move over game field by some delta.

        :param dx: Delta by x-axis.
        :param dy: Delta by y-axis.

        """
        def func(model, _gui):
            """Implement move over field."""
            model.renderer.apply_move(model, delta_x, delta_y)
        return func

    @staticmethod
    def zoom(dzoom):
        """
        Zoom game field by some delta.

        :param dzoom: Delta by which field is zoomed.

        """
        def func(model, _gui):
            """Implement field zoom."""
            model.renderer.apply_zoom(model, dzoom)
        return func

    @staticmethod
    def apply_move(bsca, *args):
        """
        Apply field move action to CA class.

        :param bsca:
            :class:`xentica.core.CellularAutomaton` instance.

        """
        for i, arg in enumerate(args):
            delta = arg
            bsca.pos[i] = (bsca.pos[i] + delta) % bsca.size[i]

    @staticmethod
    def apply_zoom(bsca, dval):
        """
        Apply field zoom action to CA class.

        :param bsca:
            :class:`xentica.core.CellularAutomaton` instance.
        :param dval:
            Delta by which field is zoomed.

        """
        bsca.zoom = max(1, (bsca.zoom + dval))

    def render_code(self):
        """Implement the code for render kernel."""
        # calculate projection plain coordinates
        code = """
            int {x} = (int) (((float) (i % width)) / (float) zoom) + dx;
            int {y} = (int) (((float) (i / width)) / (float) zoom) + dy;
            if ({x} < 0) {x} = {w}{i0} - (-{x} % {w}{i0});
            if ({x} >= {w}{i0}) {x} = {x} % {w}{i0};
            if ({y} < 0) {y} = {w}{i1} - (-{y} % {w}{i1});
            if ({y} >= {w}{i1}) {y} = {y} % {w}{i1};
            float r = 0, g = 0, b = 0;
            int num_cells_projected = 1;
        """.format(
            x="x%d" % self.projection_axes[0],
            y="x%d" % self.projection_axes[1],
            w=self.bsca.topology.lattice.width_prefix,
            i0=self.projection_axes[0],
            i1=self.projection_axes[1],
        )
        # sum over projected dimensions
        c_for = ""
        for i in range(self.bsca.topology.dimensions):
            if i in self.projection_axes:
                continue
            code += "num_cells_projected *= {w}{i};\n".format(
                i=i, w=self.bsca.topology.lattice.width_prefix
            )
            c_for += "for (int x{i} = 0; x{i} < {w}{i}; x{i}++) {{\n".format(
                i=i, w=self.bsca.topology.lattice.width_prefix
            )
        code += c_for
        code += """
            int ii = {coord_to_index};
            int3 c = col[ii];
            r += c.x;
            g += c.y;
            b += c.z;
        """.format(
            coord_to_index=self.bsca.topology.lattice.coord_to_index_code("x")
        )
        code += "}" * (self.bsca.topology.dimensions - 2)
        # calculate average
        by_smooth_factor = ""
        if self.bsca.is_constant("SMOOTH_FACTOR"):
            by_smooth_factor = "/ SMOOTH_FACTOR"
        code += """
            img[i * 3] = r / num_cells_projected {by_smooth_factor};
            img[i * 3 + 1] = g / num_cells_projected {by_smooth_factor};
            img[i * 3 + 2] = b / num_cells_projected {by_smooth_factor};
        """.format(by_smooth_factor=by_smooth_factor)
        return code



"""
The collection of classes to declare and use C variables and constants.

If the logic of your ``emit()``, ``absorb()`` or ``color()`` functions
requires the intermediate variables, you must declare them via classes
from this module in the following way::

    from xentica import core

    class MyCA(core.CellularAutomaton):
        # ...

        def emit(self):
            myvar = core.IntegerVariable()

Then you can use them in mixed expressions, like::

    myvar += self.neighbors[i].buffer.state
    self.main.state = myvar & 1

You may also define constants or other ``#define`` patterns with
:class:`Constant` class.

"""
from cached_property import cached_property

from xentica.core.mixins import BscaDetectorMixin

__all__ = ['DeferredExpression', 'Constant', 'Variable', 'IntegerVariable', ]


class DeferredExpression:
    """Base class for other classes intended to be used in mixed expressions.

    In particular, it is used in base
    :class:`Variable <xentica.core.variables.Variable>` and :class:`Property
    <xentica.core.properties.Property>` classes.

    Most of the magic methods dealing with binary and unary operators,
    as well as augmented assigns are automatically overridden for this
    class. As a result, you can use its subclasses in mixed
    expressions with ordinary Python values. See the example in
    module description above.

    Allowed binary ops
        ``+``, ``-``, ``*``, ``/``, ``%``, ``>>``, ``<<``, ``&``,
        ``^``, ``|``, ``<``, ``<=``, ``>``, ``>=``, ``==``, ``!=``

    Allowed unary ops
        ``+``, ``-``, ``~``, ``abs``, ``int``, ``float``, ``round``

    Allowed augmented assigns
        ``+=``, ``-=``, ``*=``, ``/=``, ``%=``, ``<<=``, ``>>=``,
        ``&=``, ``^=``, ``|=``

    """

    def __init__(self, code=''):
        """Override arithmetic operators and augmented assigns."""
        self.code = code
        binary_ops = (
            ('+', 'add'),
            ('-', 'sub'),
            ('*', 'mul'),
            ('/', 'truediv'),
            ('%', 'mod'),
            ('>>', 'rshift'),
            ('<<', 'lshift'),
            ('&', 'and'),
            ('^', 'xor'),
            ('|', 'or'),
            ('<', 'lt'),
            ('<=', 'le'),
            ('==', 'eq'),
            ('!=', 'ne'),
            ('>', 'gt'),
            ('>=', 'ge'),
        )
        unary_ops = (
            ('-', 'neg'),
            ('+', 'pos'),
            ('abs', 'abs'),
            ('~', 'invert'),
            ('(int)', 'int'),
            ('(float)', 'float'),
            ('round', 'round'),
        )
        for c_op, base_name in binary_ops:
            def binary_direct(oper):
                """Get direct binary operator magic method."""
                def op_func(self_var, value):
                    """Implement direct binary operator."""
                    code = "(%s %s %s)" % (self_var, oper, value)
                    return DeferredExpression(code)
                return op_func

            def binary_reflected(oper):
                """Get reflected binary operator magic method."""
                def op_func(self_var, value):
                    """Implement reflected binary operator."""
                    code = "(%s %s %s)" % (value, oper, self_var)
                    return DeferredExpression(code)
                return op_func

            def augmented_assign(oper):
                """Get augmented assign operator magic method."""
                def op_func(self_var, value):
                    """Implement augmented assign operator."""
                    if isinstance(self_var, Variable):
                        self_var.declare_once()
                    code = "%s %s= %s;\n" % (self_var.var_name, oper, value)
                    self_var.bsca.append_code(code)
                    return self
                return op_func

            func_name = "__%s__" % base_name
            setattr(self.__class__, func_name, binary_direct(c_op))
            func_name = "__r%s__" % base_name
            setattr(self.__class__, func_name, binary_reflected(c_op))
            func_name = "__i%s__" % base_name
            setattr(self.__class__, func_name, augmented_assign(c_op))

        for c_op, base_name in unary_ops:
            def unary(oper):
                """Get unary operator magic method."""
                def op_func(self_var):
                    """Implement unary operator."""
                    code = "(%s(%s))" % (oper, self_var)
                    return DeferredExpression(code)
                return op_func

            func_name = "__%s__" % base_name
            setattr(self.__class__, func_name, unary(c_op))

    def __str__(self):
        """Return the code accumulated in ``self.code``."""
        return self.code


class Constant(BscaDetectorMixin):
    """The class for defining constants and ``#define`` patterns.

    Once you instantiate :class:`Constant`, you must feed it to
    ``BSCA.define_constant()`` in order to generate correct C code::

        const = Constant("C_NAME", "bsca_var")
        self.bsca.define_constant(const)

    :param name:
        Name to use in ``#define``.
    :param value:
        String, evaluating into ``bsca.<value>``, second part of ``#define``.

    """

    def __init__(self, name, value):
        """Initialize the class."""
        self._name = name
        self._value = value
        self._pattern_name = name
        self.base_class = Constant

    def get_define_code(self):
        """Get the C code for ``#define``."""
        code = "#define %s {%s}\n" % (self._name, self._pattern_name)
        return code

    def replace_value(self, source):
        """
        Replace the constant's value in generated C code.

        :param source:
            Generated C code.

        """
        # WARNING: potentially dangerous
        val = "self._holder.%s" % self._value.split()[0]
        return source.replace("{%s}" % self._pattern_name, str(eval(val)))

    @property
    def name(self):
        """Get the name of constant."""
        return self._name


class Variable(DeferredExpression, BscaDetectorMixin):
    """
    Base class for all variables.

    Most of the functionality for variables are already implemented in
    it. Though, you are free to re-define it all to implement really
    custom behavior.

    :param val:
         Initial value for the variable.

    """

    def __init__(self, val=None):
        """Initialize base class features."""
        super(Variable, self).__init__()
        self.base_class = Variable
        self._declared = False
        if val is not None:
            self._init_val = DeferredExpression(str(val))

    @cached_property
    def var_name(self):
        """Get variable name."""
        all_vars = self._holder_frame.f_locals.items()
        for k, var in all_vars:
            if isinstance(var, self.__class__):
                if hash(self) == hash(var):
                    return k
        return "var%d" % abs(hash(self))

    def declare_once(self):
        """Declare variable and assign initial value to it."""
        if not self._declared:
            code = "%s %s = %s;\n" % (
                self.var_type, self.var_name, self._init_val
            )
            self.bsca.append_code(code)
            self._declared = True
            setattr(self.bsca, self.var_name, self)

    def __str__(self):
        """Return a variable name to use in mixed expressions."""
        return self.var_name

    def __get__(self, obj, objtype):
        """Declare a variable on first use."""
        self.declare_once()
        return self

    def __set__(self, obj, value):
        """Assign a new value to variable (doesn't work properly now)."""
        self.declare_once()
        code = "%s = %s;\n" % (self.var_name, value)
        self.bsca.append_code(code)

    @property
    def code(self):
        """Get the variable name as code."""
        return self.var_name

    @code.setter
    def code(self, val):
        """Prevent the change of code."""


class IntegerVariable(Variable):
    """The variable intended to hold a positive integer."""

    #: C type to use in definition.
    var_type = "unsigned int"

    def __init__(self, val=0):
        """Initialize variable with default value."""
        super(IntegerVariable, self).__init__(val)



"""
The collection of mixins to be used in core classes.

Would be interesting only if you are planning to hack into Xentica
core functionality.

"""
import inspect

import xentica.core.base
from xentica.core.exceptions import XenticaException

__all__ = ['BscaDetectorMixin', 'DimensionsMixin', ]


class BscaDetectorMixin:
    """
    Add a functionlality to detect BSCA class instances holding current class.

    All methods are for private use only.

    """

    @property
    def bsca(self):
        """
        Get a BSCA instance holding current class.

        Objects tree is scanned up to top and first instance found is returned.

        """
        frame = inspect.currentframe()
        while frame is not None:
            for local in frame.f_locals.values():
                if hasattr(local, "__get__"):
                    continue
                if isinstance(local, xentica.core.base.BSCA):
                    return local
            frame = frame.f_back
        raise XenticaException("BSCA not detected")

    @property
    def _holder_frame(self):
        """
        Get a frame of class instance holding current class.

        Objects tree is scanned up to top and first instance found is returned.

        """
        # As an option, we can detect base class by scanning inheritance tree:
        # inspect.getclasstree(inspect.getmro(type(self)))
        frame = inspect.currentframe().f_back.f_back.f_back
        while isinstance(frame.f_locals.get('self', ''), self.base_class):
            frame = frame.f_back
        return frame

    @property
    def _holder(self):
        """Get an instance from a frame found by :meth:`_holder_frame`."""
        return self._holder_frame.f_locals['self']


class DimensionsMixin:
    """
    The base functionality for classes, operating on a number of dimensions.

    Adds ``dimensions`` property to a class, and checks it
    automatically over a list of allowed dimensions.

    """

    #: A list of integers, containing supported dimensionality.
    #: You must set it manually for every class using :class:`DimensionsMixin`.
    supported_dimensions = []

    def __init__(self):
        """Initialize private variables."""
        self._dimensions = None

    def allowed_dimension(self, num_dim):
        """
        Test if particular dimensionality is allowed.

        :param num_dim:
            Numbers of dimensions to test

        :returns:
            Boolean value, either dimensionality is allowed or not.

        """
        return num_dim in self.supported_dimensions

    @property
    def dimensions(self):
        """Get a number of dimensions."""
        return self._dimensions

    @dimensions.setter
    def dimensions(self, num_dim):
        """Set a number of dimensions and check if it is allowed."""
        if not self.allowed_dimension(num_dim):
            msg = "%d-D %s is not supported."
            msg = msg % (num_dim, self.__class__.__name__)
            raise XenticaException(msg)
        self._dimensions = num_dim



"""
The collection of classes describing different neighborhood topologies.

All classes there are intended to be used inside ``Topology`` for
``neighborhood`` class variable definition. They are also available via
:mod:`xentica.core` shortcut. The example::

    from xentica.core import CellularAutomaton, MooreNeighborhood

    class MyCA(CellularAutomaton):
        class Topology:
            neighborhood = MooreNeighborhood()
            # ...
        # ...

"""
import itertools
import abc

from xentica.core.mixins import DimensionsMixin

__all__ = [
    'Neighborhood', 'OrthogonalNeighborhood',
    'MooreNeighborhood', 'VonNeumannNeighborhood',
]


class Neighborhood(DimensionsMixin):
    """
    Base class for all neighborhood topologies.

    For correct behavior, neighborhood classes should be inherited from
    this class. You should also implement the following functions:

    - :meth:`neighbor_coords`

    - :meth:`neighbor_state`

    See the detailed description below.

    """

    def __init__(self):
        """Initialize main attributes."""
        #: Number of neighbors, you must re-define it in sub-classes.
        self.num_neighbors = None
        #: A reference to ``Topology`` holder class, will be set in
        #: ``BSCA`` metaclass.
        self.topology = None
        self._delta2str = {-1: " - 1", 0: "", 1: " + 1"}
        self._neighbor_deltas = []
        super(Neighborhood, self).__init__()

    def __len__(self):
        """Return number of neighbors for a single cell."""
        return self.num_neighbors or 0

    @abc.abstractmethod
    def neighbor_coords(self, index, coord_prefix, neighbor_prefix):
        """
        Generate C code to obtain neighbor coordinates by its index.

        This is an abstract method, you must implement it in
        :class:`Neighborhood` subclasses.

        :param index:
            Neighbor index, a non-negative integer less than number of
            neighbors.
        :param coord_prefix:
            The prefix for variables containing main cell's coordinates.
        :param neighbor_prefix:
            The prefix for resulting variables containing neighbor coordinates.

        :returns:
            A string with C code doing all necessary to get neighbor
            state from RAM. No assignment, only a valid expression
            needed.

        """

    @abc.abstractmethod
    def neighbor_state(self, neighbor_index, state_index, coord_prefix):
        """
        Generate C code to obtain neighbor state by its index.

        This is an abstract method, you must implement it in
        :class:`Neighborhood` subclasses.

        :param neighbor_index:
            Neighbor index, a non-negative integer less than number of
            neighbors.
        :param state_index:
            State index, a non-negative integer less than number of
            neighbors for buffered states or -1 for main state.
        :param coord_prefix:
            The prefix for variables containing neighbor coordinates.

        :returns:
            A string with C code doing all necessary to process
            neighbors's coordinates and store them to neighbor
            coordinates variables.

        """


class OrthogonalNeighborhood(Neighborhood):
    """
    Base class for neighborhoods on orthogonal lattice.

    It is implementing all necessary :class:`Neighborhood` abstract
    methods, the only thing you should override is :meth:`dimensions`
    setter. In :meth:`dimensions`, you should correctly set
    ``num_neighbors`` and ``_neighbor_deltas`` attributes.

    """

    #: Any number of dimentions is supported, 100 is just to limit your
    #: hyperspatial hunger.
    supported_dimensions = list(range(1, 100))

    def neighbor_coords(self, index, coord_prefix, neighbor_prefix):
        """
        Implement neighbor coordinates obtaining by its index, in C.

        See :meth:`Neighborhood.neighbor_coords` for details.

        """
        code = ""
        for i in range(self.dimensions):
            code += "{neighbor}{i} = {coord}{i}{delta};\n".format(
                neighbor=neighbor_prefix, i=i,
                coord=coord_prefix,
                delta=self._delta2str[self._neighbor_deltas[index][i]]
            )
        return code

    def neighbor_state(self, neighbor_index, state_index, coord_prefix):
        """
        Implement state obtaining by neighbor/state index, in C.

        See :meth:`Neighborhood.neighbor_coords` for details.

        """
        cell_index = self.topology.lattice.coord_to_index_code(coord_prefix)
        if state_index + 1 > 0:
            cell_index += " + n * " + str(state_index)
        code = "fld[{cell_index}]".format(cell_index=cell_index)
        return code


class MooreNeighborhood(OrthogonalNeighborhood):
    """
    N-dimensional Moore neighborhood implementation.

    The neighbors are all cells, sharing at least one vertex.

    """

    @OrthogonalNeighborhood.dimensions.setter
    def dimensions(self, num_dim):
        """Set number of neighbors and their relative coordinates."""
        super_class = super(MooreNeighborhood, MooreNeighborhood)
        super_class.dimensions.fset(self, num_dim)
        self.num_neighbors = 3 ** num_dim - 1
        deltas = itertools.product([-1, 0, 1], repeat=num_dim)
        self._neighbor_deltas = [d for d in deltas if d != (0, 0)]


class VonNeumannNeighborhood(OrthogonalNeighborhood):
    """
    N-dimensional Von Neumann neighborhood implementation.

    The neighbors are adjacent cells in all possible orthogonal directions.

    """

    @OrthogonalNeighborhood.dimensions.setter
    def dimensions(self, num_dim):
        """Set number of neighbors and their relative coordinates."""
        super_class = super(VonNeumannNeighborhood, VonNeumannNeighborhood)
        super_class.dimensions.fset(self, num_dim)
        self.num_neighbors = 2 * num_dim
        self._neighbor_deltas = []
        for i in range(num_dim):
            delta = tuple((1 if i == j else 0 for j in range(num_dim)))
            self._neighbor_deltas.append(delta)
            delta = tuple((-1 if i == j else 0 for j in range(num_dim)))
            self._neighbor_deltas.append(delta)



"""
The module with the base class for all CA models.

All Xentica models should be inherited from :class:`CellularAutomaton`
base class. Inside the model, you should correctly define the
``Topology`` class and describe the CA logic in ``emit()``,
``absorb()`` and ``color()`` methods.

``Topology`` is the place where you define the dimensionality,
lattice, neighborhood and border effects for your CA. See
:mod:`xentica.core.topology` package for details.

The logic of the model will follow Buffered State Cellular Automaton
(BSCA) principle. In general, every cell mirrors its state in buffers
by the number of neighbors, each buffer intended for one of
neighbors. Then, at each step, the interaction between cells are
performed via buffers in 2-phase emit/absorb process. More detailed
description of BSCA principle is available in The Core section of `The
Concept`_ document.

``emit()`` describes the logic of the first phase of BSCA. At this
phase, you should fill cell's buffers with corresponding values,
depending on cell's main state and (optionally) on neighbors' main
states. The most easy logic is to just copy the main state to
buffers. It is esspecially useful when you're intending to emulate
classic CA (like Conway's Life) with BSCA. Write access to main state
is prohibited there.

``absorb()`` describes the logic of the second phase of BSCA. At this
phase, you should set the cell's main state, depending on neighbors'
buffered states. Write access to buffers is prohibited there.

``color()`` describes how to calculate cell's color from its raw
state. See detailed instructions on it in :mod:`xentica.core.color_effects`.

The logic of the functions from above will be translated into C code
at the moment of class creation. For the further instructions on how
to use cell's main and buffered states, see
:mod:`xentica.core.properties`, for the instructions on variables and
expressions with them, see :mod:`xentica.core.variables`.

A minimal example, the CA where each cell is taking the mean value of
its neighbors each step::

    from xentica import core
    from xentica.core import color_effects

    class MeanCA(core.CellularAutomaton):

        state = core.IntegerProperty(max_val=255)

        class Topology:
            dimensions = 2
            lattice = core.OrthogonalLattice()
            neighborhood = core.MooreNeighborhood()
            border = core.TorusBorder()

        def emit(self):
            for i in range(len(self.buffers)):
                self.buffers[i].state = self.main.state

        def absorb(self):
            s = core.IntegerVariable()
            for i in range(len(self.buffers)):
                s += self.neighbors[i].buffer.state
            self.main.state = s / len(self.buffers)

        @color_effects.MovingAverage
        def color(self):
            v = self.main.state
            return (v, v, v)

.. _The Concept: http://artipixoids.a5kin.net/concept/artipixoids_concept.pdf

"""
import functools
import itertools
import operator
import threading
import pickle
import collections
from copy import deepcopy

import numpy as np

from pycuda.autoinit import context
from pycuda.compiler import SourceModule
import pycuda.gpuarray as gpuarray

from xentica.bridge import MoireBridge
from xentica.seeds.random import LocalRandom
from xentica.core.properties import Property, ContainerProperty
from xentica.core.renderers import RendererPlain
from xentica.core.exceptions import XenticaException

__all__ = ['context', 'BSCA', 'CellularAutomaton', 'CachedNeighbor']


class CachedNeighbor:
    """Utility class, intended to hold main and buffered CA state."""

    def __init__(self):
        """Initialize empty main and buffered states."""
        self.main = ContainerProperty()
        self.buffer = ContainerProperty()


class BSCA(type):
    """
    Meta-class for :class:`CellularAutomaton`.

    Performs all necessary stuff to generate GPU kernels given class
    definition.

    It is also preparing ``main``, ``buffers`` and ``neighbors`` class
    variables being used in ``emit()``, ``absorb()`` and ``color()``
    methods.

    """

    mandatory_fields = (
        'dimensions', 'lattice', 'neighborhood', 'border',
    )

    @classmethod
    def __prepare__(mcs, _name, _bases):
        """Preserve the order of class variables."""
        return collections.OrderedDict()

    def __new__(mcs, name, bases, attrs):
        """Build new :class:`CellularAutomaton` class."""
        # prepare class itself
        keys = []
        for key in attrs.keys():
            if key not in ('__module__', '__qualname__'):
                keys.append(key)
        attrs['__ordered__'] = keys
        mcs._new_class = super().__new__(mcs, name, bases, attrs)
        mcs._parents = [b for b in bases if isinstance(b, BSCA)]
        if not mcs._parents:
            return mcs._new_class

        mcs._prepare_topology(mcs, attrs)

        mcs._new_class.main = None
        mcs._new_class.buffers = None
        mcs._new_class.neighbors = None
        mcs._prepare_properties(mcs, bases, attrs)

        mcs._new_class.dtype = mcs._new_class.main.dtype
        mcs._new_class.ctype = mcs._new_class.main.ctype

        mcs._new_class.constants = {}

        # set default renderer as needed
        if not hasattr(mcs._new_class, 'renderer'):
            mcs._new_class.renderer = RendererPlain()

        # build CUDA source
        source = mcs._new_class.build_emit()
        source += mcs._new_class.build_absorb()
        source += mcs._new_class.build_render()
        source = mcs._new_class.build_defines() + source
        mcs._new_class.cuda_source = source

        mcs._new_class.index_to_coord = mcs.index_to_coord
        mcs._new_class.pack_state = mcs.pack_state

        mcs._new_class.size = (1 for i in range(mcs.topology.dimensions))

        return mcs._new_class

    def _prepare_topology(cls, attrs):
        """Prepare topology for future use."""
        if hasattr(cls._new_class, 'Topology'):
            attrs['Topology'] = cls._new_class.Topology
        cls.topology = attrs.get('Topology', None)
        cls._new_class.topology = cls.topology

        if cls.topology is None:
            raise XenticaException("No Topology class declared.")

        for field in cls.mandatory_fields:
            if not hasattr(cls.topology, field):
                msg = "No %s declared in Topology class." % field
                raise XenticaException(msg)

        cls.topology.lattice.dimensions = cls.topology.dimensions
        cls.topology.neighborhood.dimensions = cls.topology.dimensions
        cls.topology.border.dimensions = cls.topology.dimensions
        cls.topology.neighborhood.topology = cls.topology
        cls.topology.border.topology = cls.topology

    def _prepare_properties(cls, bases, attrs):
        """Prepare main/buffers properties."""
        cls._new_class.main = ContainerProperty()
        cls._new_class.buffers = []
        cls._new_class.neighbors = []
        num_neighbors = len(cls.topology.neighborhood)
        for i in range(num_neighbors):
            cls._new_class.buffers.append(ContainerProperty())
            cls._new_class.neighbors.append(CachedNeighbor())
        attrs_items = [base_class.__dict__.items() for base_class in bases]
        attrs_items.append(attrs.items())
        restricted_names = {"main", "buffer"}
        for obj_name, obj in itertools.chain.from_iterable(attrs_items):
            if isinstance(obj, Property) and obj_name not in restricted_names:
                cls._new_class.main[obj_name] = deepcopy(obj)
                vname = "_cell_%s" % obj_name
                cls._new_class.main[obj_name].var_name = vname
                for i in range(num_neighbors):
                    buffers = cls._new_class.buffers
                    buffers[i][obj_name] = deepcopy(obj)
                    vname = "_bcell_%s%d" % (obj_name, i)
                    buffers[i][obj_name].var_name = vname
                    neighbor = cls._new_class.neighbors[i]
                    neighbor.main[obj_name] = deepcopy(obj)
                    vname = "_dcell_%s%d" % (obj_name, i)
                    neighbor.main[obj_name].var_name = vname
                    neighbor.buffer[obj_name] = deepcopy(obj)
                    vname = "_dbcell_%s%d" % (obj_name, i)
                    neighbor.buffer[obj_name].var_name = vname

        # propagade BSCA to properties
        cls._new_class.main.set_bsca(cls._new_class, 0, -1)
        cls._new_class.main.var_name = "_cell"
        for i in range(num_neighbors):
            cls._new_class.buffers[i].set_bsca(cls._new_class, i + 1, -1)
            cls._new_class.buffers[i].var_name = "_bcell%i" % i
            cls._new_class.neighbors[i].main.set_bsca(cls._new_class, 0, i)
            cls._new_class.neighbors[i].main.var_name = "_dcell%d" % i
            cls._new_class.neighbors[i].buffer.set_bsca(cls._new_class,
                                                        i + 1, i)
            cls._new_class.neighbors[i].buffer.var_name = "_dbcell%d" % i

    def index_to_coord(cls, i):
        """
        Wrap ``lattice.index_to_coord`` method.

        :param i: Cell's index.

        """
        return cls.topology.lattice.index_to_coord(i, cls)

    @staticmethod
    def _elementwise_kernel(name, args, body):
        """
        Build elementwise kernel using template.

        :param name:
            Kernel's name.
        :param args:
            List of kernel's arguments, consisting of
            ``("type", "var_name")`` tuples.
        :param body:
            C code for the kernel's "body", processing a single element.

        """
        arg_string = ", ".join(["%s %s" % (t, v) for t, v in args])
        kernel = """
            __global__ void %s(%s, int n) {

                unsigned tid = threadIdx.x;
                unsigned total_threads = gridDim.x * blockDim.x;
                unsigned cta_start = blockDim.x * blockIdx.x;

                for (unsigned i = cta_start + tid; i < n; i += total_threads) {
                    %s
                }

            }

        """ % (name, arg_string, body)
        return kernel

    def _translate_code(cls, *funcs):
        """
        Translate Python method to C code by execution.

        :param funcs:
            List of functions to translate in a single context.

        :returns:
            String with generated C code for elementwise kernel.

        """
        cls._func_body = ""
        cls._deferred_writes = set()
        cls._declarations = set()
        cls._unpacks = set()
        cls._coords_declared = False
        for func in funcs:
            func(cls)
        for prop in cls._deferred_writes:
            prop.deferred_write()
        return cls._func_body

    def append_code(cls, code):
        """Append ``code`` to kernel's C code."""
        cls._func_body += code

    def deferred_write(cls, prop):
        """
        Declare a property for deferred write.

        The property will be written into a memory at the end of C
        code generation.

        :param prop:
            :class:`Property <xentica.core.properties.Property>`
            subclass instance.

        """
        cls._deferred_writes.add(prop)

    def declare(cls, prop):
        """
        Mark property declared.

        :param prop:
            :class:`Property <xentica.core.properties.Property>`
            subclass instance.

        """
        cls._declarations.add(prop)

    def unpack(cls, prop):
        """
        Mark ``prop`` property unpacked.

        :param prop:
            :class:`Property <xentica.core.properties.Property>`
            subclass instance.

        """
        cls._unpacks.add(prop)

    def is_declared(cls, prop):
        """
        Check if ``prop`` property is declared.

        :param prop:
            :class:`Property <xentica.core.properties.Property>`
            subclass instance.

        :returns:
            ``True`` if property is declared, ``False`` otherwise.

        """
        return prop in cls._declarations

    def is_unpacked(cls, prop):
        """
        Check if ``prop`` property is unpacked.

        :param prop:
            :class:`Property <xentica.core.properties.Property>`
            subclass instance.

        :returns:
            ``True`` if property is unpacked, ``False`` otherwise.

        """
        return prop in cls._unpacks

    def declare_coords(cls):
        """Mark coordinate variables declared."""
        cls._coords_declared = True

    def define_constant(cls, constant):
        """
        Remember the constant is defined.

        :param constant:
            :class:`Constant <xentica.core.variables.Constant>`
            instance.

        """
        cls.constants[constant.name] = deepcopy(constant)

    def is_constant(cls, constant):
        """
        Check if the constant is defined.

        :param constant:
            :class:`Constant <xentica.core.variables.Constant>`
            instance.

        :returns:
            ``True`` if constant is defined, ``False`` otherwise.

        """
        return constant in cls.constants

    @property
    def coords_declared(cls):
        """
        Check if coordinate variables are declared.

        :returns:
            ``True`` if coordinate variables are declared, ``False``
            otherwise.

        """
        return cls._coords_declared

    def build_defines(cls):
        """
        Generate ``#define`` section for all kernels.

        :returns:
            String with C code with necessary defines.

        """
        defines = ""
        for const in cls.constants.values():
            defines += const.get_define_code()
        return defines

    def build_emit(cls):
        """
        Generate ``emit()`` kernel.

        :returns:
            String with C code for ``emit()`` kernel.

        """
        args = [(cls.ctype, "*fld"), ]
        body = cls._translate_code(cls.emit)
        return cls._elementwise_kernel("emit", args, body)

    def build_absorb(cls):
        """
        Generate ``absorb()`` kernel.

        :returns:
            String with C code for ``absorb()`` kernel.

        """
        args = [(cls.ctype, "*fld"), ("int3", "*col")]
        body = cls._translate_code(cls.absorb, cls.color)
        return cls._elementwise_kernel("absorb", args, body)

    def build_render(cls):
        """
        Generate ``render()`` kernel.

        :returns:
            String with C code for ``render()`` kernel.

        """
        args = cls.renderer.args
        body = cls.renderer.render_code()
        return cls._elementwise_kernel("render", args, body)

    def pack_state(cls, state):
        """
        Pack state structure into raw in-memory representation.

        :returns:
            Integer representing packed state.

        """
        val = 0
        shift = 0
        for name, prop in cls.main.properties.items():
            val += state[name] << shift
            shift += prop.bit_width
        return val


class GPUKernels:
    """Class incapsulating GPU kernels."""

    def __init__(self, source):
        """
        Initialize GPU kernels.

        :param source:
            CUDA module source code in C.

        """
        cuda_module = SourceModule(source)
        self.emit = cuda_module.get_function("emit")
        self.absorb = cuda_module.get_function("absorb")
        self.render = cuda_module.get_function("render")


class GPUArrays:
    """Class incapsulating GPU arrays."""

    def __init__(self, init_cells, init_colors):
        """
        Initialize GPU arrays.

        :param init_cells:
            NumPy array to put to cells' GPU array.
        :param init_colors:
            NumPy array to put to colors' GPU array.

        """
        self.img = None
        self.colors = gpuarray.to_gpu(init_colors)
        self.cells = gpuarray.to_gpu(init_cells)

    def init_img(self, num_cells):
        """
        Initilatize image for rendering.

        :param num_cells:
            Total number of cells.

        """
        self.img = gpuarray.zeros((num_cells, ), dtype=np.int32)


class GPU:
    """Holder class, incapsulating GPU kernels and arrays."""

    def __init__(self, source, init_cells, init_colors):
        """
        Initialize GPU arrays and kernels.

        :param source:
            CUDA module source code in C.
        :param init_cells:
            NumPy array to put to cells' GPU array.
        :param init_colors:
            NumPy array to put to colors' GPU array.

        """
        self.kernels = GPUKernels(source)
        self.arrays = GPUArrays(init_cells, init_colors)


class CellularAutomaton(metaclass=BSCA):
    """
    Base class for all Xentica models.

    Compiles GPU kernels generated by :class:`BSCA` metaclass,
    initializes necessary GPU arrays and popupates them with the seed.

    After initialization, you can run step-by-step simulation and
    render the field at any moment::

        from xentica import core
        import moire

        class MyCA(core.CellularAutomaton):
            # ...

        class MyExperiment(core.Experiment):
            # ...

        ca = MyCA(MyExperiment)
        ca.set_viewport((320, 200))

        # run CA manually for 100 steps
        for i in range(100):
            ca.step()
        # render current timestep
        frame = ca.render()

        # or run the whole process interactively with Moire
        gui = moire.GUI(runnable=ca)
        gui.run()

    :param experiment_class:
        :class:`Experiment <xentica.core.experiments.Experiment>`
        instance, holding all necessary parameters for the field
        initialization.

    """

    def __init__(self, experiment_class):
        """Initialize kernels, GPU arrays and other stuff."""
        # visuals
        self.frame_buf = np.zeros((3, ), dtype=np.uint8)
        self.width, self.height = 0, 0
        # populate attributes from Experiment class
        for attr_name in dir(experiment_class):
            attr = getattr(experiment_class, attr_name)
            if (not callable(attr) and not attr_name.startswith("__")):
                if attr_name == 'seed':
                    continue
                setattr(self, attr_name, attr)
        # default simulation values
        self.speed = 1
        self.paused = False
        self.timestep = 0
        # CUDA kernel
        self.cells_num = functools.reduce(operator.mul, self.size)
        source = self.cuda_source
        for const in self.constants.values():
            source = const.replace_value(source)
        # print(source)
        # build seed
        init_colors = np.zeros((self.cells_num * 3, ), dtype=np.int32)
        cells_total = self.cells_num * (len(self.buffers) + 1) + 1
        self.random = LocalRandom(experiment_class.word)
        experiment_class.seed.random = self.random
        init_cells = np.zeros((cells_total, ), dtype=self.dtype)
        experiment_class.seed.generate(init_cells, self)
        # initialize GPU stuff
        self.gpu = GPU(source, init_cells, init_colors)
        # bridge
        self.bridge = MoireBridge
        self.renderer.setup_actions(self.bridge)
        # lock
        self._lock = threading.Lock()

    def apply_speed(self, dval):
        """
        Change the simulation speed.

        Usable only in conduction with Moire, although you can use the
        ``speed`` value in your custom GUI too.

        :param dval: Delta by which speed is changed.

        """
        self.speed = max(1, (self.speed + dval))

    def toggle_pause(self):
        """
        Toggle ``paused`` flag.

        When paused, the ``step()`` method does nothing.

        """
        self.paused = not self.paused

    def set_viewport(self, size):
        """
        Set viewport (camera) size and initialize GPU array for it.

        :param size: tuple with width and height in pixels.

        """
        self.width, self.height = size
        num_cells = self.width * self.height * 3
        self.gpu.arrays.init_img(num_cells)

    def step(self):
        """
        Perform a single simulation step.

        ``timestep`` attribute will hold the current step number.

        """
        if self.paused:
            return
        # pylint: disable=protected-access
        # This is "hack" to get block/grid sizes, it's vital to us.
        # No way to get it correctly with PyCuda right now.
        block, grid = self.gpu.arrays.cells._block, self.gpu.arrays.cells._grid
        with self._lock:
            self.gpu.kernels.emit(self.gpu.arrays.cells,
                                  np.int32(self.cells_num),
                                  block=block, grid=grid)
            self.gpu.kernels.absorb(self.gpu.arrays.cells,
                                    self.gpu.arrays.colors,
                                    np.int32(self.cells_num),
                                    block=block, grid=grid)
            self.timestep += 1

    def render(self):
        """
        Render the field at the current timestep.

        You must call :meth:`set_viewport` before do any rendering.

        :returns:
            NumPy array of ``np.uint8`` values, ``width * height * 3``
            size. The RGB values are consecutive.

        """
        # pylint: disable=protected-access
        # This is "hack" to get block/grid sizes, it's vital to us.
        # No way to get it correctly with PyCuda right now.
        if self.gpu.arrays.img is None:
            msg = "Viewport is not set, call set_viewport() before rendering."
            raise XenticaException(msg)
        block, grid = self.gpu.arrays.img._block, self.gpu.arrays.img._grid
        with self._lock:
            args = self.renderer.get_args_vals(self)
            args.append(np.int32(self.width * self.height))
            self.gpu.kernels.render(*args, block=block, grid=grid)
            return self.gpu.arrays.img.get().astype(np.uint8)

    def save(self, filename):
        """Save the CA state into ``filename`` file."""
        with open(filename, "wb") as ca_file:
            ca_state = {
                "cells": self.gpu.arrays.cells.get(),
                "colors": self.gpu.arrays.colors.get(),
                "random": self.random,
            }
            pickle.dump(ca_state, ca_file)

    def load(self, filename):
        """Load the CA state from ``filename`` file."""
        with open(filename, "rb") as ca_file:
            ca_state = pickle.load(ca_file)
            self.gpu.arrays.cells = gpuarray.to_gpu(ca_state['cells'])
            self.gpu.arrays.colors = gpuarray.to_gpu(ca_state['colors'])
            self.random.load(ca_state['random'])



"""
The collection of decorators for ``color()`` method, each CA model should have.

The method should be decorated by one of the classes below, otherwise
the correct model behavior is not guaranteed.

All decorators are get the ``(red, green, blue)`` tuple from
``color()`` method, then process it to create some color effect.

The minimal example::

    from xentica import core
    from xentica.core import color_effects

    class MyCA(core.CellularAutomaton):

        state = core.IntegerProperty(max_val=1)

        # ...

        @color_effects.MovingAverage
        def color(self):
            red = self.main.state * 255
            green = self.main.state * 255
            blue = self.main.state * 255
            return (red, green, blue)

"""
from xentica.core.variables import Constant
from xentica.core.mixins import BscaDetectorMixin

__all__ = ['ColorEffect', 'MovingAverage', ]


class ColorEffect(BscaDetectorMixin):
    """
    Base class for other color effects.

    You may also use it as standalone color effect decorator, it just
    doing nothing, storing calculated RGB value directly.

    To create your own class inherited from :class:`ColorEffect`, you
    should override ``__call__`` method, and place a code of color
    processing into ``self.effect``. The code should process a values
    of ``new_r``, ``new_g``, ``new_b`` variables and store the result
    back to them.

    The example::

        class MyEffect(ColorEffect):

            def __call__(self, *args):
                self.effect = "new_r += 20;"
                self.effect += "new_g += 15;"
                self.effect += "new_b += 10;"
                return super(MyEffect, self).__call__(*args)

    """

    def __init__(self, func):
        """Initialize base attributes."""
        self.func = func
        self.effect = ""

    def __call__(self, self_var):
        """
        Implement the color decorator.

        Sibling classes should override this method, and return
        ``super`` result, like shown in the example above.

        """
        red, green, blue = self.func(self_var)
        code = """
            int new_r = %s;
            int new_g = %s;
            int new_b = %s;
            %s
            col[i] = make_int3(new_r, new_g, new_b);
        """ % (red, green, blue, self.effect)
        self_var.append_code(code)


class MovingAverage(ColorEffect):
    """
    Apply the moving average to each color channel separately.

    With this effect, 3 additional settings are available for you in
    ``Experiment`` classes:

    fade_in
        The maximum delta by which a channel could
        *increase* its value in a single timestep.

    fade_out
        The maximum delta by which a channel could
        *decrease* its value in a single timestep.

    smooth_factor
        The divisor for two previous settings, to make
        the effect even smoother.

    """

    def __call__(self, self_var):
        """Implement the effect."""
        self.bsca.define_constant(Constant("FADE_IN", "fade_in"))
        self.bsca.define_constant(Constant("FADE_OUT", "fade_out"))
        self.bsca.define_constant(Constant("SMOOTH_FACTOR",
                                           "smooth_factor"))
        self.bsca.fade_in = 255
        self.bsca.fade_out = 255
        self.bsca.smooth_factor = 1
        self.effect = """
            new_r *= SMOOTH_FACTOR;
            new_g *= SMOOTH_FACTOR;
            new_b *= SMOOTH_FACTOR;
            int3 old_col = col[i];
            new_r = max(min(new_r, old_col.x + FADE_IN),
                        old_col.x - FADE_OUT);
            new_g = max(min(new_g, old_col.y + FADE_IN),
                        old_col.y - FADE_OUT);
            new_b = max(min(new_b, old_col.z + FADE_IN),
                        old_col.z - FADE_OUT);
        """
        return super(MovingAverage, self).__call__(self_var)



"""
The collection of classes describing different types of field borders.

All classes there are intended to be used inside ``Topology`` for
``border`` class variable definition. They are also available via
:mod:`xentica.core` shortcut. The example::

    from xentica.core import CellularAutomaton, TorusBorder

    class MyCA(CellularAutomaton):
        class Topology:
            border = TorusBorder()
            # ...
        # ...

"""
import abc

from xentica.core.mixins import DimensionsMixin

__all__ = [
    'Border', 'WrappedBorder', 'GeneratedBorder',
    'TorusBorder', 'StaticBorder',
]


class Border(DimensionsMixin):
    """
    Base class for all types of borders.

    You should not inherit your borders directly from this class, use
    either :class:`WrappedBorder` or :class:`GeneratedBorder` base
    subclasses for this.

    """

    def __init__(self):
        """Initialize common things for all borders."""
        self.topology = None
        super(Border, self).__init__()


class WrappedBorder(Border):
    """
    Base class for borders wrapping the field into different manifolds.

    For correct behavior, you should implement :meth:`wrap_coords` method.

    See the detailed description below.

    """

    @abc.abstractmethod
    def wrap_coords(self, coord_prefix):
        """
        Generate C code to translate off-board coordinates to on-board ones.

        This is an abstract method, you must implement it in
        :class:`WrappedBorder` subclasses.

        :param coord_prefix:
            The prefix for variables containing cell's coordinates.

        """


class GeneratedBorder(Border):
    """
    Base class for borders generating states of the off-board cells.

    For correct behavior, you should implement :meth:`off_board_state` method.

    See the detailed description below.

    """

    @abc.abstractmethod
    def off_board_state(self, coord_prefix):
        """
        Generate C code to obtain off-board cell's state.

        This is an abstract method, you must implement it in
        :class:`GeneratedBorder` subclasses.

        :param coord_prefix:
            The prefix for variables containing cell's coordinates.

        """


class TorusBorder(WrappedBorder):
    """
    Wraps the entire field into N-torus manifold.

    This is the most common type of border, allowing you to generate
    seamless tiles for wallpapers.

    """

    #: Any number of dimentions is supported, 100 is just to limit your
    #: hyperspatial hunger.
    supported_dimensions = list(range(1, 100))

    def wrap_coords(self, coord_prefix):
        """
        Impement coordinates wrapping to torus.

        See :meth:`WrappedBorder.wrap_coords` for details.

        """
        code = ""
        for i in range(self.dimensions):
            code += "{x}{i} = ({x}{i} + {w}{i}) % {w}{i};\n".format(
                x=coord_prefix, i=i,
                w=self.topology.lattice.width_prefix
            )
        return code


class StaticBorder(GeneratedBorder):
    """
    Generates a static value for every off-board cell.

    This is acting like your field is surrounded by cells with the
    same pre-defined state.

    The default is just an empty (zero) state.

    """

    supported_dimensions = list(range(1, 100))

    def __init__(self, value=0):
        """
        Store the static value.

        :param value:
            A static value to be used for every off-board cell.

        """
        self._value = value
        super(StaticBorder, self).__init__()

    def off_board_state(self, coord_prefix):
        """
        Impement off-board cells' values obtaining.

        See :meth:`GeneratedBorder.off_board_state` for details.

        """
        return str(self._value)



"""
This module contains the main class to be used in custom bridges.

Methods from :class:`Bridge` class should be used in other bridges

"""

__all__ = ['Bridge', ]


class Bridge:
    """Main bridge class containing basic functions."""

    @staticmethod
    def noop(_model, _gui):
        """Do nothing."""

    @staticmethod
    def exit_app(_model, gui):
        """Exit GUI application."""
        gui.exit_app()

    @staticmethod
    def speed(dspeed):
        """Change simulation speed."""
        def func(model, _gui):
            """Apply speed."""
            model.apply_speed(dspeed)
        return func

    @staticmethod
    def toggle_pause(model, _gui):
        """Pause/unpause simulation."""
        model.toggle_pause()

    @staticmethod
    def toggle_sysinfo(_model, gui):
        """Turn system info panel on/off."""
        gui.sysinfo.toggle()



"""
Module with the bridge to `Moire`_ UI.

.. _Moire: https://github.com/a5kin/moire

"""
from xentica.bridge.base import Bridge

__all__ = ['MoireBridge', ]


class MoireBridge:
    """
    Class incaplulating the actions for Moire UI.

    ``[`` Speed simulation down.

    ``]`` Speed simulation up.

    ``SPACEBAR`` Pause/unpause simulation.

    ``F12`` Toggle system info.

    ``ESC`` Exit app.

    """

    key_actions = {
        "[": Bridge.speed(-1),
        "]": Bridge.speed(1),
        "spacebar": Bridge.toggle_pause,
        "f12": Bridge.toggle_sysinfo,
        "escape": Bridge.exit_app,
    }



# -*- coding: utf-8 -*-
"""
Xentica core functionality is available via modules from this package.

In addition, you may use ``core`` package as a shortcut to the main
classes of the framework.

- **Base classes**
   - ``core.CellularAutomaton`` →
     :class:`xentica.core.base.CellularAutomaton`
   - ``core.Experiment`` →
     :class:`xentica.core.experiment.Experiment`

- **Lattices**
   - ``core.OrthogonalLattice`` →
     :class:`xentica.core.topology.lattice.OrthogonalLattice`

- **Neighborhoods**
   - ``core.MooreNeighborhood`` →
     :class:`xentica.core.topology.neighborhood.MooreNeighborhood`
   - ``core.VonNeumannNeighborhood`` →
     :class:`xentica.core.topology.neighborhood.VonNeumannNeighborhood`

- **Borders**
   - ``core.TorusBorder`` →
     :class:`xentica.core.topology.border.TorusBorder`
   - ``core.StaticBorder`` →
     :class:`xentica.core.topology.border.StaticBorder`

- **Properties**
   - ``core.IntegerProperty`` →
     :class:`xentica.core.properties.IntegerProperty`

- **Variables**
   - ``core.IntegerVariable`` →
     :class:`xentica.core.variables.IntegerVariable`

The classes listed above are all you need to build CA models and
experiments with Xentica, unless you are planning to implement custom
core features like new lattices, borders, etc.

"""
from xentica.core.base import CellularAutomaton
from xentica.core.properties import (
    IntegerProperty,
)
from xentica.core.variables import (
    IntegerVariable,
)
from xentica.core.topology.lattice import (
    OrthogonalLattice,
)
from xentica.core.topology.neighborhood import (
    MooreNeighborhood, VonNeumannNeighborhood
)
from xentica.core.topology.border import (
    TorusBorder, StaticBorder,
)
from xentica.core.experiment import Experiment

__all__ = [
    'CellularAutomaton',
    'IntegerProperty',
    'IntegerVariable',
    'OrthogonalLattice',
    'MooreNeighborhood',
    'VonNeumannNeighborhood',
    'TorusBorder',
    'StaticBorder',
    'Experiment',
]



"""
The bridge between Xentica and GUI interface.

This package contains all necessary stuff to connect Xentica framework
to custom interactive visualization environments.

Right now, only one environment (`Moire`_) is available. This is the
official environment, evolving along with the main framework. You are
free to implement your own environments. If so, please make a PR on
Github and we'll include your solution to the bridge.

Bridge functions are automatically used when you run the simulation
like this::

    import moire
    ca = MyCellularAutomaton(MyExperiment)
    gui = moire.GUI(runnable=ca)
    gui.run()

.. _Moire: https://github.com/a5kin/moire

"""
from xentica.bridge.moire import MoireBridge


__all__ = [
    "MoireBridge",
]



"""
The package helping you to construct initial CA state (seed).

Classes from modules below are intended for use in
:class:`Experiment <xentica.core.experiments.Experiment>` classes.

For example to initialize the whole board with random values::

    from xentica import core, seeds

    class MyExperiment(core.Experiment):
        # ...
        seed = seeds.patterns.PrimordialSoup(
            vals={
                "state": seeds.random.RandInt(0, 1),
            }
       )

"""

from xentica.seeds import random
from xentica.seeds import patterns

__all__ = ['random', 'patterns', ]



"""
The module for package-wide RNG.

The main intention is to keep separate deterministic random streams
for every :class:`CellularAutomaton<xentica.core.base.CellularAutomaton>`
instance. So, is you're initialized RNG for a particular CA with some
seed, you're get the guarantee that the random sequence will be the
same, no matter how many other CA's you're running in parallel.

"""
import random
import functools
import operator

import numpy as np

__all__ = ['LocalRandom', 'RandInt', ]


class LocalRandom:
    """
    The holder class for the RNG sequence.

    It is incapsulating both standart Python random stream and NumPy one.

    Once instantiated, you can use them as follows::

        from xentica.seeds.random import LocalRandom

        random = LocalRandom()
        # get random number from standard stream
        val = random.standard.randint(1, 10)
        # get 100 random numbers from NumPy stream
        vals = random.numpy.randint(1, 10, 100)

    """

    def __init__(self, seed=None):
        """Initialize local random streams."""
        self.standard = random.Random(seed)
        np_seed = self.standard.getrandbits(32)
        self.numpy = np.random.RandomState(np_seed)

    def load(self, rng):
        """
        Load random state from the class.

        :param rng: :class:`LocalRandom` instance.

        """
        self.standard = rng.standard
        self.numpy = rng.numpy


class RandInt:
    """
    Class, generating a sequence of random integers in some interval.

    It is intended to be used in
    :class:`Experiment <xentica.core.experiments.Experiment>`
    seeds. See the example of initializing CA property above.

    :param min_val: Lower bound for random value.
    :param max_val: Upper bound for random value.

    """

    def __init__(self, min_val, max_val):
        """Initialize the random sequence."""
        self.min_val = min_val
        self.max_val = max_val

    def __get__(self, instance, owner):
        """
        Get the random value in specified range from standard stream.

        This method is used automatically from
        :class:`CellularAutomaton<xentica.core.base.CellularAutomaton>`,
        at the stage of constructing the initial board state.

        """
        if hasattr(instance, "size"):
            num_values = functools.reduce(operator.mul, instance.size)
            return instance.random.numpy.randint(self.min_val,
                                                 self.max_val + 1,
                                                 num_values)
        return instance.random.standard.randint(self.min_val, self.max_val)



"""
The collection of classes to describe properties of CA models.

.. warning::
    Do not confuse with Python properties.

Xentica properties are declaring as class variables and helping you to
organize CA state into complex structures.

Each :class:`CellularAutomaton <xentica.core.base.CellularAutomaton>`
instance should have at least one property declared. The property name
is up to you. If your model has just one value for state (like in most
classic CA), the best practice is to call it ``state`` as follows::

    from xentica import core

    class MyCA(core.CellularAutomaton):
        state = core.IntegerProperty(max_val=1)
        # ...

Then, you can use it in expressions of ``emit()``, ``absorb()`` and
``color()`` functions as:

``self.main.state``
    to get and set main state;

``self.buffers[i].state``
    to get and set i-th buffered state;

``self.neighbors[i].buffer.state``
    to get and set i-th neighbor buffered state.

Xentica will take care of all other things, like packing CA properties
into binary representation and back, storing and getting corresponding
values from VRAM, etc.

Most of properties will return
:class:`DeferredExpression <xentica.core.variables.DeferredExpression>`
on access, so you can use them safely in mixed expressions::

    self.buffers[i].state = self.main.state + 1

"""
import math
import abc
from collections import OrderedDict

import numpy as np
from cached_property import cached_property

from xentica.core.variables import DeferredExpression

__all__ = ['Property', 'IntegerProperty', 'ContainerProperty', ]


class Property(DeferredExpression):
    """
    Base class for all properties.

    Has a vast set of default functionality already
    implemented. Though, you are free to re-define it all to implement
    really custom behavior.

    """

    def __init__(self):
        """Initialize default attributes."""
        self.bsca = None
        self._types = (
            # (bit_width, numpy_dtype, gpu_c_type)
            (8, np.uint8, 'char'),
            (16, np.uint16, 'short'),
            (32, np.uint32, 'int'),
        )
        self._buf_num, self._nbr_num = 1, 1
        super(Property, self).__init__()

    @cached_property
    def best_type(self):
        """
        Get type that suits best to store a property.

        :returns:
            tuple representing best type:
            ``(bit_width, numpy_dtype, gpu_c_type)``

        """
        _best_type = self._types[-1]
        for _type in self._types:
            type_width = _type[0]
            if self.bit_width <= type_width:
                _best_type = _type
                break
        return _best_type

    @cached_property
    def dtype(self):
        """
        Get NumPy dtype, based on result of :meth:`best_type`.

        :returns:
            NumPy dtype that suits best to store a property.

        """
        return self.best_type[1]

    @cached_property
    def ctype(self):
        """
        Get C type, based on result of :meth:`best_type`.

        :returns:
            C type that suits best to store a property.

        """
        return 'unsigned ' + self.best_type[2]

    @cached_property
    def bit_width(self):
        """
        Get the number of bits necessary to store a property.

        :returns:
            Positive integer, a property's bit width.

        """
        return self.calc_bit_width()

    @cached_property
    def width(self):
        """
        Get the number of memory cells to store a property.

        In example, if ``ctype == "int"`` and ``bit_width == 64``, you
        need 2 memory cells.

        :returns:
            Positive integer, a property's width.

        """
        type_width = self.best_type[0]
        return int(math.ceil(self.bit_width / type_width))

    @abc.abstractmethod
    def calc_bit_width(self):
        """
        Calculate the property's bit width.

        This is the method you most likely need to override. It will
        be called from :meth:`bit_width`.

        :returns:
            Positive integer, calculated property's width in bits.

        """
        return 1  # default, just for consistency

    def set_bsca(self, bsca, buf_num, nbr_num):
        """
        Set up a reference to BSCA instance.

        Do not override this method, it is cruicial to inner framework
        mechanics.

        :param bsca:
            :class:`CellularAutomaton <xentica.core.base.CellularAutomaton>`
            instance.
        :param buf_num:
            Buffer's index, associated to property.
        :param nbr_num:
            Neighbor's index, associated to property.

        """
        self.bsca = bsca
        self._buf_num = buf_num
        self._nbr_num = nbr_num

    def __getattribute__(self, attr):
        """Implement custom logic when property is get as class attribute."""
        obj = object.__getattribute__(self, attr)
        if hasattr(obj, '__get__') and attr != "__class__":
            return obj.__get__(self, type(self))
        return obj

    def __setattr__(self, attr, val):
        """Implement custom logic when property is set as class attribute."""
        try:
            obj = object.__getattribute__(self, attr)
        except AttributeError:
            object.__setattr__(self, attr, val)
        else:
            if hasattr(obj, '__set__'):
                obj.__set__(self, val)
            else:
                object.__setattr__(self, attr, val)

    def __get__(self, obj, objtype):
        """Implement custom logic when property is get as class descriptor."""
        self.declare_once()
        return DeferredExpression(self.var_name)

    def __set__(self, obj, value):
        """Implement custom logic when property is set as class descriptor."""
        self.declare_once()
        if not hasattr(value, "code"):
            value = DeferredExpression(str(value))
        code = "%s = %s;\n" % (self.var_name, value.code)
        self.bsca.append_code(code)

    @cached_property
    def _mem_cell(self):
        """
        Generate C expression to get cell's state from RAM.

        :returns:
            String with C expression getting the state from memory.

        """
        if self._nbr_num >= 0:
            neighborhood = self.bsca.topology.neighborhood
            return neighborhood.neighbor_state(self._nbr_num,
                                               self._buf_num, "_nx")
        offset = ""
        if self._buf_num > 0:
            offset = " + n * %d" % self._buf_num
        return "fld[i%s]" % offset

    @property
    def declared(self):
        """Test if the state variable is declared."""
        if self.bsca is None:
            return False
        return self.bsca.is_declared(self)

    @property
    def coords_declared(self):
        """Test if the coordinates variables are declared."""
        if self.bsca is None:
            return True
        return self.bsca.coords_declared

    def declare_once(self):
        """
        Generate C code to declare a variable holding cell's state.

        You must push the generated code to BSCA via
        ``self.bsca.append_code()``, then declare necessary stuff via
        ``self.bsca.declare()``.

        You should also take care of skipping the whole process if
        things are already declared.

        """
        if self.declared:
            return
        code = "%s %s;\n" % (self.ctype, self.var_name)
        self.bsca.append_code(code)
        self.bsca.declare(self)


class IntegerProperty(Property):
    """
    Most generic property for you to use.

    It is just a positive integer with upper limit of ``max_val``.

    """

    def __init__(self, max_val):
        """Initialize class specific attributes."""
        self.max_val = max_val
        self._buf_num = 0
        super(IntegerProperty, self).__init__()

    def calc_bit_width(self):
        """Calculate bit width, based on ``max_val``."""
        return int(math.log2(self.max_val)) + 1


class ContainerProperty(Property):
    """
    A property acting as a holder for other properties.

    Currently is used only for inner framework mechanics, in
    particular, to hold, pack and unpack all top-level properties.

    It will be enhanced in future versions, and give you the
    ability to implement nested properties structures.

    .. warning::
        Right now, direct use of this class is prohibited.

    """

    def __init__(self):
        """Initialize ``OrderedDict`` to hold other properties."""
        super(ContainerProperty, self).__init__()
        self._properties = OrderedDict()
        self.init_val = None

    @property
    def properties(self):
        """Get inner properties."""
        return self._properties

    def values(self):
        """Iterate over properties, emulating ``dict`` functionality."""
        for prop in self._properties.values():
            yield prop

    @property
    def unpacked(self):
        """Test if inner properties are unpacked from memory."""
        if self.bsca is None:
            return False
        return self.bsca.is_unpacked(self)

    def __getitem__(self, key):
        """Get property by key, emulating ``dict`` functionality."""
        return self._properties[key]

    def __setitem__(self, key, val):
        """Set property by key, emulating ``dict`` functionality."""
        self._properties[key] = val
        object.__setattr__(self, key, val)

    def calc_bit_width(self):
        """Calculate bit width as sum of inner properties' bit widths."""
        return sum([p.bit_width for p in self._properties.values()])

    def set_bsca(self, bsca, buf_num, nbr_num):
        """Propagate BSCA setting to inner properties."""
        self.bsca = bsca
        self._buf_num = buf_num
        self._nbr_num = nbr_num
        for key in self._properties.keys():
            self[key].set_bsca(bsca, buf_num, nbr_num)

    def __get__(self, obj, objtype):
        """Return self reference when getting as class descriptor."""
        return self

    def __set__(self, obj, value):
        """Do nothing when setting as class descriptor."""
        pass

    def __getattribute__(self, attr):
        """Get value from VRAM and unpack it to variables."""
        obj = object.__getattribute__(self, attr)
        if isinstance(obj, Property):
            self.init_val = self._mem_cell
            self.declare_once()
            self._unpack_state()
            return obj.__get__(self, type(self))
        return obj

    def __setattr__(self, attr, val):
        """Declare resulting variable and defer the write memory access."""
        try:
            obj = object.__getattribute__(self, attr)
        except AttributeError:
            object.__setattr__(self, attr, val)
        else:
            if isinstance(obj, Property):
                obj.declare_once()
                obj.__set__(self, val)
                self.declare_once()
                self.bsca.deferred_write(self)
                self.bsca.unpack(self)
            else:
                object.__setattr__(self, attr, val)

    def declare_once(self):
        """
        Do all necessary declarations for inner properties.

        Also, implements the case of off-board neighbor access.

        :param init_val:
             Default value for the property.

        """
        if self.declared:
            return
        code = ""
        if self._nbr_num >= 0:
            neighborhood = self.bsca.topology.neighborhood
            lattice = self.bsca.topology.lattice
            border = self.bsca.topology.border
            dimensions = self.bsca.topology.dimensions

            if not self.coords_declared:
                code += lattice.index_to_coord_code("i", "_x")
                coord_vars = ["_nx%d" % i for i in range(dimensions)]
                code += "int %s;\n" % ", ".join(coord_vars)
                self.bsca.declare_coords()

            code += neighborhood.neighbor_coords(self._nbr_num, "_x", "_nx")
            is_cell_off_board = lattice.is_off_board_code("_nx")
            if hasattr(border, "wrap_coords"):
                code += """
                    if ({is_cell_off_board}) {{
                        {wrap_coords}
                    }}
                """.format(
                    is_cell_off_board=is_cell_off_board,
                    wrap_coords=border.wrap_coords("_nx"),
                )
            else:
                code += """
                    {type} {var};
                    if ({is_cell_off_board}) {{
                        {var} = {off_board_cell};
                    }} else {{
                        {var} = {neighbor_state};
                    }}
                """.format(
                    type=self.ctype, var=self.var_name,
                    is_cell_off_board=is_cell_off_board,
                    off_board_cell=border.off_board_state("_nx"),
                    neighbor_state=self.init_val,
                )
                self.bsca.append_code(code)
                self.bsca.declare(self)
                return

        if self.init_val is None:
            code += "%s %s;\n" % (
                self.ctype, self.var_name
            )
        else:
            code += "%s %s = %s;\n" % (
                self.ctype, self.var_name, self.init_val
            )
        self.bsca.append_code(code)
        self.bsca.declare(self)

    def _unpack_state(self):
        """Unpack inner properties values from in-memory representation."""
        if self.unpacked:
            return
        code = ""
        shift = 0
        for prop in self._properties.values():
            prop.declare_once()
            val = self.var_name
            if shift > 0:
                val += " >> %d" % shift
            mask = 2 ** prop.bit_width - 1
            val = "({val}) & {mask}".format(val=val, mask=mask)
            code += "{var} = {val};\n".format(var=prop.var_name, val=val)
            shift += prop.bit_width
        self.bsca.append_code(code)
        self.bsca.unpack(self)

    def deferred_write(self):
        """
        Pack state and write its value to VRAM.

        This method is called from ``BSCA`` at the end of kernel processing.

        """
        shift = 0
        vals = []
        for prop in self._properties.values():
            prop.declare_once()
            mask = 2 ** prop.bit_width - 1
            val = "({val} & {mask})".format(val=prop.var_name, mask=mask)
            if shift > 0:
                val = "({val} << {shift})".format(val=val, shift=shift)
            vals.append(val)
            shift += prop.bit_width
        summed_vals = " + ".join(vals)
        code = "{var} = {val};\n".format(var=self.var_name, val=summed_vals)
        code += "%s = %s;\n" % (self._mem_cell, self.var_name)
        self.bsca.append_code(code)



"""
The collection of classes describing different lattice topologies.

All classes there are intended to be used inside ``Topology`` for
``lattice`` class variable definition. They are also available via
:mod:`xentica.core` shortcut. The example::

    from xentica.core import CellularAutomaton, OrthogonalLattice

    class MyCA(CellularAutomaton):
        class Topology:
            lattice = OrthogonalLattice()
            # ...
        # ...

"""
import abc

from xentica.core.mixins import DimensionsMixin

from xentica.core.mixins import BscaDetectorMixin
from xentica.core.variables import Constant

__all__ = ['Lattice', 'OrthogonalLattice', ]


class Lattice(DimensionsMixin, BscaDetectorMixin, metaclass=abc.ABCMeta):
    """
    Base class for all lattices.

    For correct behavior, lattice classes should be inherited from
    this class. You should also implement the following functions:

    - :meth:`index_to_coord_code`

    - :meth:`index_to_coord`

    - :meth:`coord_to_index_code`

    - :meth:`is_off_board_code`

    See the detailed description below.

    """

    #: The prefix to be used in C code for field size constants.
    width_prefix = "_w"

    def _define_constants_once(self):
        """Define field size conctants in C code."""
        for i in range(self.bsca.topology.dimensions):
            constant = Constant("%s%d" % (self.width_prefix, i),
                                "size[%d]" % i)
            self.bsca.define_constant(constant)

    @abc.abstractmethod
    def index_to_coord_code(self, index_name, coord_prefix):
        """
        Generate C code to obtain coordinates by cell's index.

        This is an abstract method, you must implement it in :class:`Lattice`
        subclasses.

        :param index_name:
            The name of variable containing cell's index.
        :param coord_prefix:
            The prefix for resulting variables, containing coordinates.

        :returns:
            A string with C code, doing all necessary to process
            index variable and store coordinates to variables with
            given prefix.

        """

    @abc.abstractmethod
    def index_to_coord(self, idx, bsca):
        """Obtain cell's coordinates by its index, in pure Python.

        This is an abstract method, you must implement it in :class:`Lattice`
        subclasses.

        :param idx:
            Cell's index, a positive integer, or NumPy array of indices.
        :param bsca:
            :class:`xentica.core.CellularAutomaton` instance, to access
            field size and number of dimensions.

        :returns:
            Tuple of integer coordinates, or NumPy arrays of coords
            per each axis.

        """

    @abc.abstractmethod
    def coord_to_index_code(self, coord_prefix):
        """
        Generate C code for obtaining cell's index by coordinates.

        This is an abstract method, you must implement it in :class:`Lattice`
        subclasses.

        :param coord_prefix:
            The prefix for variables, containing coordinates.

        :returns:
            A string with C code calculating cell's index. No
            assignment, only a valid expression needed.

        """

    @abc.abstractmethod
    def is_off_board_code(self, coord_prefix):
        """
        Generate C code to test if the cell's coordinates are off board.

        This is an abstract method, you must implement it in :class:`Lattice`
        subclasses.

        :param coord_prefix:
            The prefix for variables, containing coordinates.

        :returns:
            A string with C code testing coordinate variables. No
            assignment, only a valid expression with boolean result needed.

        """


class OrthogonalLattice(Lattice):
    """
    N-dimensional orthogonal lattice.

    Points are all possible positive integer coordinates.

    """

    #: Overridden value for supported dimensions.
    supported_dimensions = list(range(1, 100))

    def index_to_coord_code(self, index_name, coord_prefix):
        """
        Implement coordinates obtaining by cell's index in C.

        See :meth:`Lattice.index_to_coord_code` for details.

        """
        self._define_constants_once()
        i = 0

        def wrap_format(text):
            """Format helper."""
            return text.format(x=coord_prefix, i=i,
                               index=index_name, w=self.width_prefix)

        for i in range(self.dimensions):
            if i == 0:
                code = wrap_format("int _{index} = {index};\n")
                index_name = "_" + index_name
            if i < self.dimensions - 1:
                code += wrap_format("int {x}{i} = {index} % {w}{i};\n")
                code += wrap_format("{index} /= {w}{i};\n")
            else:
                code += wrap_format("int {x}{i} = {index};\n")
        return code

    def index_to_coord(self, idx, bsca):
        """
        Implement coordinates obtaining by cell's index in Python.

        See :meth:`Lattice.index_to_coord` for details.

        """
        coord = []
        for i in range(bsca.topology.dimensions):
            if i < self.dimensions - 1:
                x_i = idx % bsca.size[i]
                idx //= bsca.size[i]
            else:
                x_i = idx
            coord.append(x_i)
        return coord

    def coord_to_index_code(self, coord_prefix):
        """
        Implement cell's index obtaining by coordinates in C.

        See :meth:`Lattice.coord_to_index_code` for details.

        """
        self._define_constants_once()

        summands = []
        for i in range(self.dimensions):
            summand = coord_prefix + str(i)
            for j in range(i):
                summand = self.width_prefix + str(j) + " * " + summand
            summands.append(summand)
        return " + ".join(summands)

    def is_off_board_code(self, coord_prefix):
        """
        Implement off board cell obtaining in C.

        See :meth:`Lattice.is_off_board_code` for details.

        """
        self._define_constants_once()

        conditions = []
        for i in range(self.dimensions):
            condition = "{x}{i} < 0 || {x}{i} >= {w}{i}".format(
                x=coord_prefix, i=i, w=self.width_prefix
            )
            conditions.append(condition)
        return " || ".join(conditions)



"""This package helps you build the topology for CA models.

All :class:`xentica.core.CellularAutomaton` subclasses **must** have
``Topology`` class declared inside. This class describes:

- ``dimensions``: the number of dimensions your CA model operates on.

- ``lattice``: the type of lattice of your CA board. Built-in lattice
  types are available in :mod:`xentica.core.topology.lattice` module.

- ``neighborhood``: the type of neighborhood for a single
  cell. Built-in neighborhood types are available in
  :mod:`xentica.core.topology.neighborhood` module.

- ``border``: the type of border effect, e.g. how to process off-board
  cells. Built-in border types are available in
  :mod:`xentica.core.topology.border` module.

In example, you can declare the topology for a 2-dimensional
orthogonal lattice with Moore neighborhood, wrapped to a 3-torus, as
follows::

    class Topology:

        dimensions = 2
        lattice = core.OrthogonalLattice()
        neighborhood = core.MooreNeighborhood()
        border = core.TorusBorder()

"""



"""
A setuptools based setup module.

Adopted from:
https://github.com/pypa/sampleproject

"""

import codecs
from os import path

from setuptools import setup, find_packages


def get_long_description():
    """Load description from README file."""
    here = path.abspath(path.dirname(__file__))
    with codecs.open(path.join(here, 'README.rst'), encoding='utf-8') as file:
        return file.read()


setup(
    name='xentica',
    version='0.1.1',
    description='GPU-accelerated engine for multi-dimensional '
                'cellular automata',
    long_description=get_long_description(),
    url='https://github.com/a5kin/xentica',
    author='Andrey Zamaraev',
    author_email='a5kin@github.com',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Games/Entertainment :: Simulation',
        'Topic :: Scientific/Engineering :: Artificial Life',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Software Development :: Libraries',
        'Topic :: Artistic Software',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    keywords='cellular automata ca simulation artificial life alife energy '
             'framework gpu multidimensional microprograms conservation',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[
        'numpy',
        'cached-property',
    ],
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
)



"""
Module containing different patterns for CA seed initialization.

Each pattern class having one mandatory method ``generate()`` which is
called automatically at the initialization stage.

Patterns are intended to use in
:class:`Experiment <xentica.core.experiments.Experiment>` classes.
See the example of general usage above.

.. _The Concept: http://artipixoids.a5kin.net/concept/artipixoids_concept.pdf

"""

import abc

import numpy as np

from .random import LocalRandom

__all__ = ['RandomPattern', 'BigBang', 'PrimordialSoup', 'ValDict', ]


class ValDictMeta(type):
    """Placeholder for :class:`ValDict` metaclass."""


class ValDict(metaclass=ValDictMeta):
    """
    Wrapper over Python dictionary.

    It can keep descriptor classes along with regular values. Then, on
    the item getting, the necessary value is automatically obtaining
    either directly or via descriptor logic.

    Readonly, you should set all dictionary values at the class
    initialization.

    Example of usage::

        >>> from xentica.seeds.random import RandInt
        >>> from xentica.seeds.patterns import ValDict
        >>> d = {'a': 2, 's': RandInt(11, 23), 'd': 3.3}
        >>> vd = ValDict(d)
        >>> vd['a']
        2
        >>> vd['s']
        14
        >>> vd['d']
        3.3

    :param d:
        Dictionary with mixed values. May contain descriptor classes.
    :param parent:
        A reference to class holding the dictionary. Optional.

    """

    def __init__(self, d, parent=None):
        """Initialize class."""
        self._d = d
        self.parent = parent
        if parent is None:
            self.parent = self
            self.random = LocalRandom()

    def items(self):
        """Iterate over dictionary items."""
        for key in self._d.keys():
            val = self[key]
            yield key, val

    def keys(self):
        """Iterate over dictionary keys."""
        for k in self._d.keys():
            yield k

    def __getitem__(self, key):
        """Implement the logic of obtaining item from dictionary."""
        if key in self._d:
            if hasattr(self._d[key], '__get__'):
                return self._d[key].__get__(self.parent, self.parent.__class__)
            return self._d[key]
        raise KeyError(key)

    def __setitem__(self, key, val):
        """Suppress direct item setting, may be allowed in future."""
        raise NotImplementedError


class RandomPattern:
    """
    Base class for random patterns.

    :param vals:
        Dictionary with mixed values. May contain descriptor classes.

    """

    def __init__(self, vals):
        """Initialize class."""
        self.random = LocalRandom()
        self.vals = ValDict(vals, self)

    @abc.abstractmethod
    def generate(self, cells, bsca):
        """
        Generate the entire initial state.

        This is an abstract method, you must implement it in
        :class:`RandomPattern` subclasses.

        :param cells:
            NumPy array with cells' states as items. The seed will be
            generated over this array.
        :param cells_num:
            Total number of cells in ``cells`` array.
        :param field_size:
            Tuple with field sizes per each dimension.
        :param index_to_coord:
            Function translating cell's index to coordinate.
        :param pack_state:
            Function packing state into single integer.

        """


class BigBang(RandomPattern):
    """
    Random init pattern, known as *"Big Bang"*.

    Citation from `The Concept`_:

        *"A small area of space is initialized with a high amount of energy
        and random parameters per each quantum. Outside the area, quanta
        has either zero or minimum possible amount of energy. This is a
        good test for the ability of energy to spread in empty space."*

    The current implementation allows to generate a value for every
    cell inside specified N-cube area. Cells outside the area have
    zero values.

    :param vals:
        Dictionary with mixed values. May contain descriptor classes.
    :param pos:
        A tuple with the coordinates of the lowest corner of the Bang area.
    :param size:
        A tuple with the size of Bang area per each dimension.

    """

    def __init__(self, vals, pos=None, size=None):
        """Initialize class."""
        self.pos = np.asarray(pos) if pos else None
        self.size = np.asarray(size) if size else None
        super(BigBang, self).__init__(vals)

    def _prepare_area(self, bsca_size):
        """
        Prepare area size and position.

        :param bsca_size: tuple with CA size.

        """
        dims = range(len(bsca_size))
        randint = self.random.standard.randint
        if self.size is None:
            rnd_vec = [randint(1, bsca_size[i]) for i in dims]
            self.size = np.asarray(rnd_vec)
        if self.pos is None:
            rnd_vec = [randint(0, bsca_size[i]) for i in dims]
            self.pos = np.asarray(rnd_vec)
        for i in range(len(self.pos)):
            coord, width, side = self.pos[i], self.size[i], bsca_size[i]
            if coord + width >= side:
                self.pos[i] = side - width
            self.pos[i] = max(0, self.pos[i])

    def generate(self, cells, bsca):
        """
        Generate the entire initial state.

        See :meth:`RandomPattern.generate` for details.

        """
        self._prepare_area(bsca.size)
        indices = np.arange(0, bsca.cells_num)
        coords = bsca.index_to_coord(indices)
        area = None
        for i in range(len(bsca.size)):
            condition = (coords[i] >= self.pos[i])
            condition &= (coords[i] < self.pos[i] + self.size[i])
            if area is None:
                area = condition
                continue
            area &= (condition)
        state = {}
        for name in sorted(self.vals.keys()):
            val = self.vals[name]
            state[name] = val
        cells[np.where(area)] = bsca.pack_state(state)


class PrimordialSoup(RandomPattern):
    """
    Random init pattern, known as *"Primordial Soup"*.

    Citation from `The Concept`_:

        *"Each and every quantum initially has an equally small amount
        of energy, other parameters are random. This is a good test
        for the ability of energy to self-organize in clusters from
        the completely uniform distribution."*

    The current implementation allows to populate the entire board
    with generated values.

    :param vals:
        Dictionary with mixed values. May contain descriptor classes.

    """

    def __init__(self, vals):
        """Initialize class."""
        self.size = None
        super(PrimordialSoup, self).__init__(vals)

    def generate(self, cells, bsca):
        """
        Generate the entire initial state.

        See :meth:`RandomPattern.generate` for details.

        """
        self.size = bsca.size
        state = {}
        for name in sorted(self.vals.keys()):
            val = self.vals[name]
            state[name] = val
        cells[:bsca.cells_num] = bsca.pack_state(state)



"""The collection of formatters."""


def sizeof_fmt(num, suffix=''):
    """
    Format the number with the humanized order of magnitude.

    In example, ``11234`` become ``11.23K``.

    :param num:
        The positive integer.
    :param suffix:
        Additional suffix to add to formatted string.

    :returns:
        Formatted number as a string.

    """
    if abs(num) < 1000:
        return "%s%s" % (num, suffix)
    for unit in ['K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        num /= 1000.0
        if abs(num) < 1000.0:
            return "%.2f%s%s" % (num, unit, suffix)
    num /= 1000.0
    return "%.2f%s%s" % (num, 'Y', suffix)



"""
A collection of models derived from Conway's Game Of Life.

Experiment classes included.

"""
from xentica import core
from xentica import seeds
from xentica.core import color_effects


class GameOfLife(core.CellularAutomaton):
    """
    The classic CA built with Xentica framework.

    It has only one property called ``state``, which is positive
    integer with max value of 1.

    """

    state = core.IntegerProperty(max_val=1)

    class Topology:
        """
        Mandatory class for all ``CellularAutomaton`` instances.

        All class variables below are also mandatory.

        Here, we declare the topology as a 2-dimensional orthogonal
        lattice with Moore neighborhood, wrapped to a 3-torus.

        """

        dimensions = 2
        lattice = core.OrthogonalLattice()
        neighborhood = core.MooreNeighborhood()
        border = core.TorusBorder()

    def emit(self):
        """
        Implement the logic of emit phase.

        Statements below will be translated into C code as emit kernel
        at the moment of class creation.

        Here, we just copy main state to surrounding buffers.

        """
        for i in range(len(self.buffers)):
            self.buffers[i].state = self.main.state

    def absorb(self):
        """
        Implement the logic of absorb phase.

        Statements below will be translated into C code as well.

        Here, we sum all neigbors buffered states and apply Conway
        rule to modify cell's own state.

        """
        neighbors_alive = core.IntegerVariable()
        for i in range(len(self.buffers)):
            neighbors_alive += self.neighbors[i].buffer.state
        is_born = (8 >> neighbors_alive) & 1
        is_sustain = (12 >> neighbors_alive) & 1
        self.main.state = is_born | is_sustain & self.main.state

    @color_effects.MovingAverage
    def color(self):
        """
        Implement the logic of cell's color calculation.

        Must return a tuple of RGB values computed from ``self.main``
        properties.

        Also, must be decorated by a class from ``color_effects``
        module.

        Here, we simply define 0 state as pure black, and 1 state as
        pure white.

        """
        red = self.main.state * 255
        green = self.main.state * 255
        blue = self.main.state * 255
        return (red, green, blue)


class GameOfLifeStatic(GameOfLife):
    """
    Game of Life variant with static border made of live cells.

    This is an example of how easy you can inherit other models.

    """

    class Topology(GameOfLife.Topology):
        """
        You can inherit parent class ``Topology``.

        Then, override only necessary variables.

        """

        border = core.StaticBorder(1)


class GameOfLifeColor(GameOfLife):
    """
    Game Of Life variant with RGB color.

    This is an example of how to use multiple properties per cell.

    """

    state = core.IntegerProperty(max_val=1)
    red = core.IntegerProperty(max_val=255)
    green = core.IntegerProperty(max_val=255)
    blue = core.IntegerProperty(max_val=255)

    def emit(self):
        """Copy all properties to surrounding buffers."""
        for i in range(len(self.buffers)):
            self.buffers[i].state = self.main.state
            self.buffers[i].red = self.main.red
            self.buffers[i].green = self.main.green
            self.buffers[i].blue = self.main.blue

    def absorb(self):
        """
        Calculate RGB as neighbors sum for living cell only.

        Note, parent ``absorb`` method should be called using direct
        class access, not via ``super``.

        """
        GameOfLife.absorb(self)
        red_sum = core.IntegerVariable()
        for i in range(len(self.buffers)):
            red_sum += self.neighbors[i].buffer.red + 1
        green_sum = core.IntegerVariable()
        for i in range(len(self.buffers)):
            green_sum += self.neighbors[i].buffer.green + 1
        blue_sum = core.IntegerVariable()
        for i in range(len(self.buffers)):
            blue_sum += self.neighbors[i].buffer.blue + 1
        self.main.red = red_sum * self.main.state
        self.main.green = green_sum * self.main.state
        self.main.blue = blue_sum * self.main.state

    @color_effects.MovingAverage
    def color(self):
        """Calculate color as usual."""
        red = self.main.state * self.main.red
        green = self.main.state * self.main.green
        blue = self.main.state * self.main.blue
        return (red, green, blue)


class GameOfLife6D(GameOfLife):
    """
    Game of Life variant in 6D.

    Nothing interesting, just to prove you can do it with ease.

    """

    class Topology(GameOfLife.Topology):
        """
        Hyper-spacewalk, is as easy as increase ``dimensions`` value.

        However, we are also changing neighborhood to Von Neumann
        here, to prevent neighbors number exponential grow.

        """

        dimensions = 6
        neighborhood = core.VonNeumannNeighborhood()


class GOLExperiment(core.Experiment):
    """
    Particular experiment for the vanilla Game of Life.

    Here, we define constants and initial conditions from which the
    world's seed will be generated.

    The ``word`` is RNG seed string. The ``size``, ``zoom`` and
    ``pos`` are board contstants. The ``seed`` is a pattern used in
    initial board state generation.

    ``BigBang`` is a pattern when small area initialized with a
    high-density random values.

    """

    word = "OBEY XENTICA"
    size = (640, 360, )
    zoom = 3
    pos = [0, 0]
    seed = seeds.patterns.BigBang(
        pos=(320, 180),
        size=(100, 100),
        vals={
            "state": seeds.random.RandInt(0, 1),
        }
    )


class GOLExperiment2(GOLExperiment):
    """
    Another experiment for the vanilla GoL.

    Since it is inherited from ``GOLExperiment``, we can redefine only
    values we need.

    ``PrimordialSoup`` is a pattern when the whole board is
    initialized with low-density random values.

    """

    word = "XENTICA IS YOUR GODDESS"
    seed = seeds.patterns.PrimordialSoup(
        vals={
            "state": seeds.random.RandInt(0, 1),
        }
    )


class GOLExperimentColor(GOLExperiment):
    """
    Experiment for GameOfLifeColor.

    Here, we introduce ``fade_out`` constant, which is used in
    rendering causing cells slowly fade out.

    Note, it is only aestetic effect, and does not affect real cell
    state.

    """

    fade_out = 10
    seed = seeds.patterns.PrimordialSoup(
        vals={
            "state": seeds.random.RandInt(0, 1),
            "red": seeds.random.RandInt(0, 255),
            "green": seeds.random.RandInt(0, 255),
            "blue": seeds.random.RandInt(0, 255),
        }
    )


class GOLExperiment6D(GOLExperiment2):
    """
    Special experiment for 6D Life.

    Here, we define the world with 2 spatial and 4 looped
    micro-dimensions, 3 cells length each.

    As a result, we get large quasi-stable oscillators, looping over
    micro-dimensions. Strangely formed, but nothing interesting,
    really.

    """

    size = (640, 360, 3, 3, 3, 3)


def main():
    """Run model/experiment interactively."""
    import moire
    model = GameOfLifeColor(GOLExperimentColor)
    gui = moire.GUI(runnable=model)
    gui.run()


if __name__ == "__main__":
    main()
